package org.openxma.dsl.platform.dao.impl;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.Criteria;
import org.hibernate.Hibernate;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.SessionFactory;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Example;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Restrictions;
import org.hibernate.engine.SessionFactoryImplementor;
import org.hibernate.impl.CriteriaImpl;
import org.hibernate.type.Type;
import org.openxma.dsl.platform.dao.FilterExpression;
import org.openxma.dsl.platform.dao.FinderArgumentTypeFactory;
import org.openxma.dsl.platform.dao.FinderExecutor;
import org.openxma.dsl.platform.dao.FinderNamingStrategy;
import org.openxma.dsl.platform.dao.GenericDao;
import org.openxma.dsl.platform.dao.SortOrder;
import org.openxma.dsl.platform.dao.SortOrderSpecification;
import org.openxma.dsl.platform.hibernate.NamedQueryFilter;
import org.openxma.dsl.platform.query.QueryObject;
import org.openxma.dsl.platform.service.Mapper;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.jdbc.core.simple.SimpleJdbcCall;
import org.springframework.jdbc.datasource.SingleConnectionDataSource;
import org.springframework.jdbc.support.incrementer.DataFieldMaxValueIncrementer;
import org.springframework.orm.hibernate3.SessionFactoryUtils;
import org.springframework.util.Assert;

/**
 * Hibernate implementation of GenericDao A typesafe implementation of CRUD and finder methods based on Hibernate and
 * Spring AOP The finders are implemented through the executeFinder method. Normally called by the
 * FinderIntroductionInterceptor
 */
public class GenericDaoHibernateImpl<T, Pk extends Serializable> implements GenericDao<T, Pk>, FinderExecutor<T>,
        InitializingBean {
    public static final String QUERY_NAME = "queryName";
    /** Logger available to subclasses */
    protected final Log logger = LogFactory.getLog(getClass());
    protected DataFieldMaxValueIncrementer dataFieldMaxValueIncrementer;
    protected SessionFactory sessionFactory;
    protected DataSource dataSource;
    protected FinderNamingStrategy namingStrategy = new SimpleFinderNamingStrategy();
    protected FinderArgumentTypeFactory argumentTypeFactory = new SimpleFinderArgumentTypeFactory();
    protected final Class<? extends T> type;

    /**
     * @param sessionFactory
     *            the sessionFactory to set
     */
    @Autowired
    public void setSessionFactory(@Qualifier("default") SessionFactory sessionFactory) {
        this.sessionFactory = sessionFactory;
    }

    /**
     * @param dataFieldMaxValueIncrementer
     *            the dataFieldMaxValueIncrementer to set
     */
    @Autowired
    public void setDataFieldMaxValueIncrementer(
            @Qualifier("default") DataFieldMaxValueIncrementer dataFieldMaxValueIncrementer) {
        this.dataFieldMaxValueIncrementer = dataFieldMaxValueIncrementer;
    }

    /**
     * @param dataSource
     *            the dataSource to set
     */
    @Autowired(required=false)
    public void setDataSource(@Qualifier("default") DataSource dataSource) {
        this.dataSource = dataSource;
    }

    /**
     * @param finderNamingStrategy
     *            the finderNamingStrategy to set
     */
    public void setNamingStrategy(FinderNamingStrategy finderNamingStrategy) {
        this.namingStrategy = finderNamingStrategy;
    }

    /**
     * @param finderArgumentTypeFactory
     *            the finderArgumentTypeFactory to set
     */
    public void setArgumentTypeFactory(FinderArgumentTypeFactory finderArgumentTypeFactory) {
        this.argumentTypeFactory = finderArgumentTypeFactory;
    }

    public GenericDaoHibernateImpl(Class<? extends T> type) {
        this.type = type;
    }

    /**
     * @return a new instance of the configured entity Type
     */
    protected T createEntityInstance() {
        return (T) BeanUtils.instantiateClass(type);
    }

    /**
     * {@inheritDoc}
     */
    public T read(Pk id) {
        return (T) this.sessionFactory.getCurrentSession().get(type, id);
    }

    /**
     * {@inheritDoc}
     */
    public void update(T o) {
        this.sessionFactory.getCurrentSession().update(o);
    }

    /**
     * {@inheritDoc}
     */
    public void delete(T o) {
        this.sessionFactory.getCurrentSession().delete(o);
    }

    /**
     * {@inheritDoc}
     */
    public void delete(Pk oid) {
        delete(load(oid));
    }

    /**
     * {@inheritDoc}
     */
    public Set<T> findAll() {
        return new LinkedHashSet<T>(this.sessionFactory.getCurrentSession().createCriteria(type).list());
    }

    /**
     * {@inheritDoc}
     */
    public Set<T> findAll(SortOrderSpecification sortOrderSpecification) {
        Criteria criteria = this.sessionFactory.getCurrentSession().createCriteria(type);
        applySortOrder(sortOrderSpecification, criteria);
        return new LinkedHashSet<T>(criteria.list());
    }

    /**
     * {@inheritDoc}
     */
    public List<T> findByExample(T example) {
        Criteria exmapleCriteria = this.sessionFactory.getCurrentSession().createCriteria(type);
        return exmapleCriteria.add(Example.create(example)).list();
    }

    /**
     * {@inheritDoc}
     */
    public Set<T> find(FilterExpression filterExpression, SortOrderSpecification orderSpecification) {
        Assert.notNull(filterExpression, "Parameter 'filterExpression' must not be null");
        Criteria criteria = this.sessionFactory.getCurrentSession().createCriteria(type);
        Criterion criterion = (Criterion) filterExpression.applyTo(new HibernateFilter(criteria));
        criteria.add(criterion);
        if (null != orderSpecification) {
            applySortOrder(orderSpecification, criteria);
        }
        return new LinkedHashSet(criteria.list());
    }

    /**
     * {@inheritDoc}
     */
    public T get(Pk oid) {
        return (T) this.sessionFactory.getCurrentSession().get(type, oid);
    }

    /**
     * {@inheritDoc}
     */
    public T load(Pk oid) {
        T entity = (T) this.sessionFactory.getCurrentSession().load(type, oid);
        // force eager initialization
        Hibernate.initialize(entity);
        return entity;
    }

    /**
     * {@inheritDoc}
     */
    public T merge(T entity) {
        return (T) this.sessionFactory.getCurrentSession().merge(entity);
    }

    /**
     * {@inheritDoc}
     */
    public void saveOrUpdate(T entity) {
        this.sessionFactory.getCurrentSession().saveOrUpdate(entity);
    }

    /**
     * {@inheritDoc}
     */
    public List<T> executeFinder(Method method, final Object[] queryArgs) {
        final Query namedQuery = prepareQuery(method, queryArgs);
        return namedQuery.list();
    }

    /**
     * {@inheritDoc}
     */
    public Iterator<T> iterateFinder(Method method, final Object[] queryArgs) {
        final Query namedQuery = prepareQuery(method, queryArgs);
        return namedQuery.iterate();
    }

    /**
     * {@inheritDoc}
     */
    public void refresh(T object) {
        this.sessionFactory.getCurrentSession().refresh(object);
    }

    /**
     * {@inheritDoc}
     */
    public void evict(T object) {
        this.sessionFactory.getCurrentSession().evict(object);
    }

    /**
     * {@inheritDoc}
     */
    public <E> Collection<E> query(QueryObject queryObject) {
        if (queryObject.getQueryName() == null) {
            throw new IllegalArgumentException("parameter 'queryObject.queryName' must not be null");
        }
        NamedQueryFilter namedQueryContext = createQueryContext(queryObject);
        return (Collection) queryObject.evaluate(namedQueryContext);
    }

    protected NamedQueryFilter createQueryContext(QueryObject queryObject) {
        return new NamedQueryFilter(this.sessionFactory, getNamedQueryMapper(queryObject.getQueryName()));
    }

    protected Criteria createCriteria(Class entityClass) {
        return this.sessionFactory.getCurrentSession().createCriteria(entityClass);
    }

    protected Criteria createCriteria() {
        return createCriteria(type);
    }

    protected Criteria createCriteria(Criterion criterion) {
        Criteria criteria = createCriteria(type);
        if (criterion != null) {
            criteria = add(criteria, criterion);
        }
        return criteria;
    }

    protected List<T> list(Criteria criteria) {
        if (((CriteriaImpl) criteria).iterateExpressionEntries().hasNext()) {
            return criteria.list();
        }
        return Collections.emptyList();
    }

    protected T unique(Criteria criteria) {
        if (((CriteriaImpl) criteria).iterateExpressionEntries().hasNext()) {
            return (T) criteria.uniqueResult();
        }
        return null;
    }

    protected <E> List<E> list(Query query) {
        return query.list();
    }

    protected <S> S unique(Query query) {
        return (S) query.uniqueResult();
    }

    protected void orderBy(Criteria criteria, SortOrderSpecification sortOrderSpecification) {
        for (Iterator<SortOrder> iterator = sortOrderSpecification.getSortOrders().iterator(); iterator.hasNext();) {
            SortOrder sortOrder = (SortOrder) iterator.next();
            criteria.addOrder(sortOrder.isAscending() ? Order.asc(sortOrder.getPropertyName()) : Order.desc(sortOrder
                    .getPropertyName()));
        }
    }

    protected Criterion and(Criterion criterion, String propertyName, Object value, boolean isNullable) {
        Criterion expression = equal(propertyName, value, isNullable);
        return null == criterion ? expression : (null == expression ? criterion : Restrictions.and(criterion,
                expression));
    }

    protected Criterion or(Criterion criterion, String propertyName, Object value, boolean isNullable) {
        Criterion expression = equal(propertyName, value, isNullable);
        return null == criterion ? expression : (null == expression ? criterion : Restrictions
                .or(criterion, expression));
    }

    protected Criterion equal(String propertyName, Object value, boolean isNullable) {
        return null != value ? Restrictions.eq(propertyName, value) : (isNullable ? Restrictions.isNull(propertyName)
                : null);
    }

    protected Criteria add(Criteria criteria, Criterion criterion) {
        if (criterion != null) {
            criteria.add(criterion);
        }
        return criteria;
    }

    protected Criteria join(Criteria criteria, String reference, FilterExpression filterExpression) {
        if (filterExpression != null) {
            Criterion expression = (Criterion) filterExpression.applyTo(new HibernateFilter(criteria, reference));
            criteria.add(expression);
        }
        return criteria;
    }

    protected final void applyPagingParameters(final Criteria criteria, final Integer firstResult,
            final Integer maxResults) {
        if (null != firstResult) {
            criteria.setFirstResult(firstResult.intValue());
        }
        if (null != maxResults && maxResults.intValue() > 0) {
            criteria.setMaxResults(maxResults.intValue());
        }
    }

    protected Mapper getNamedQueryMapper(String queryName) {
        return new NamedQueryMapper();
    }

    protected SimpleJdbcCall createJdbcCall(String name) {
        return createJdbcCall(name,null);
    }
    
    protected SimpleJdbcCall createJdbcCall(String name,String catalog) {
        Assert.notNull(name, "Parameter 'name' must not be null");
        SimpleJdbcCall simpleJdbcCall = new SimpleJdbcCall(createJdbcTemplate());
        simpleJdbcCall.setProcedureName(name);
        simpleJdbcCall.setCatalogName(catalog);
        return simpleJdbcCall;
    }
    
    protected Map<String,Object> executeJdbcCall(SimpleJdbcCall jdbcCall, Map<String,Object> arguments) {
        Assert.notNull(jdbcCall, "Parameter 'jdbcCall' must not be null");
        SqlParameterSource sqlParameterSource = new MapSqlParameterSource(arguments);
        return executeJdbcCall(jdbcCall, sqlParameterSource);
    }

    protected Map<String,Object> executeJdbcCall(SimpleJdbcCall jdbcCall, SqlParameterSource sqlParameterSource) {
        Assert.notNull(jdbcCall, "Parameter 'jdbcCall' must not be null");
        return jdbcCall.execute(sqlParameterSource);
    }

    protected JdbcTemplate createJdbcTemplate() {
        if (this.dataSource == null) {
            SessionFactoryImplementor sessionFactoryImplementor = (SessionFactoryImplementor) this.sessionFactory;
            try {
                return new JdbcTemplate(new SingleConnectionDataSource(sessionFactoryImplementor
                        .getConnectionProvider().getConnection(), false));
            } catch (SQLException e) {
                throw SessionFactoryUtils.convertHibernateAccessException(new HibernateException(e));
            }
        }
        return new JdbcTemplate(this.dataSource);
    }

    protected void applySortOrder(SortOrderSpecification orderSpecification, Criteria criteria) {
        Assert.notNull(orderSpecification, "Parameter 'orderSpecification' must not be null");
        for (Iterator<SortOrder> iterator = orderSpecification.getSortOrders().iterator(); iterator.hasNext();) {
            SortOrder order = iterator.next();
            criteria.addOrder(order.isAscending() ? org.hibernate.criterion.Order.asc(order.getPropertyName())
                    : org.hibernate.criterion.Order.desc(order.getPropertyName()));
        }
    }

    protected Query prepareQuery(Method method, Object[] queryArgs) {
        final String queryName = namingStrategy.queryNameFromMethod(type, method);
        final Query namedQuery = this.sessionFactory.getCurrentSession().getNamedQuery(queryName);
        String[] namedParameters = namedQuery.getNamedParameters();
        if (namedParameters.length == 0) {
            setPositionalParams(queryArgs, namedQuery);
        } else {
            setNamedParams(namedParameters, queryArgs, namedQuery);
        }
        return namedQuery;
    }

    protected void setPositionalParams(Object[] queryArgs, Query namedQuery) {
        // Set parameter. Use custom Hibernate Type if necessary
        if (queryArgs != null) {
            for (int i = 0; i < queryArgs.length; i++) {
                Object arg = queryArgs[i];
                Type argType = this.argumentTypeFactory.getArgumentType(arg);
                if (argType != null) {
                    namedQuery.setParameter(i, arg, argType);
                } else {
                    namedQuery.setParameter(i, arg);
                }
            }
        }
    }

    protected void setNamedParams(String[] namedParameters, Object[] queryArgs, Query namedQuery) {
        // Set parameter. Use custom Hibernate Type if necessary
        if (queryArgs != null) {
            for (int i = 0; i < queryArgs.length; i++) {
                Object arg = queryArgs[i];
                Type argType = this.argumentTypeFactory.getArgumentType(arg);
                if (argType != null) {
                    namedQuery.setParameter(namedParameters[i], arg, argType);
                } else {
                    if (arg instanceof Collection) {
                        namedQuery.setParameterList(namedParameters[i], (Collection) arg);
                    } else {
                        namedQuery.setParameter(namedParameters[i], arg);
                    }
                }
            }
        }
    }

    protected class NamedQueryMapper implements Mapper {

        public <E> List<E> createAndMapMany(Collection<?> sourceCollection, Class<E> targetClass) {
            throw new UnsupportedOperationException();
        }

        public <E> List<E> createAndMapMany(Collection<?> sourceCollection, Class<E> targetClass, Object context) {
            int rowNumber = 0;
            List<E> resultList = new ArrayList<E>();
            Map<Object, Object> contextMap = getContextAsMap(context);
            for (Object object : sourceCollection) {
                Object[] tuple = (Object[]) (object instanceof Object[] ? object : new Object[] { object });
                contextMap.put("rowNumber", rowNumber);
                E result = createAndMapOne(tuple, targetClass, context);
                if (result != null) {
                    resultList.add(result);
                }
            }
            return resultList;
        }

        public <E> E createAndMapOne(Object sourceObject, Class<E> targetClass) {
            throw new UnsupportedOperationException();
        }

        public <E> E createAndMapOne(Object sourceObject, Class<E> targetClass, Object context) {
            E newInstance = null;
            if (targetClass != null) {
                newInstance = (E) BeanUtils.instantiateClass(targetClass);
            }
            return mapOne(sourceObject, newInstance, context);

        }

        public <E> E mapOne(Object sourceObject, E targetObject) {
            throw new UnsupportedOperationException();
        }

        public <E> E mapOne(Object sourceObject, E targetObject, Object context) {
            return (E) sourceObject;
        }

        protected String getQueryName(Object context) {
            return (String) getContextAsMap(context).get(QUERY_NAME);
        }

        protected Map<Object, Object> getContextAsMap(Object context) {
            Map<Object, Object> contextAsMap = (Map<Object, Object>) context;
            return new HashMap<Object, Object>(contextAsMap);
        }

    }

    public void afterPropertiesSet() throws Exception {
        if (this.dataSource == null) {
            logger.warn("Missing configuration for default qualified 'dataSource' property.");
        }
    }
}
